#include <windows.h>
#include <Fl/gl.h>
#include <gl/glu.h>
#include <stdio.h>

#include "camera.h"

#pragma warning(push)
#pragma warning(disable : 4244)

#ifndef M_PI
#define M_PI 3.141592653589793238462643383279502f
#endif 

const float kMouseRotationSensitivity		= 1.0f/90.0f;
const float kMouseTranslationXSensitivity	= 0.03f;
const float kMouseTranslationYSensitivity	= 0.03f;
const float kMouseZoomSensitivity			= 0.08f;



void Camera::calculateViewingTransformParameters() 
{
	Mat4f dollyXform;
	Mat4f azimXform;
	Mat4f elevXform;
	Mat4f twistXform;
	Mat4f originXform;

	Vec3f upVector;

	makeHTrans(dollyXform, Vec3f(0,0,mDolly));
	makeHRotY(azimXform, mAzimuth);
	makeHRotX(elevXform, mElevation);
	makeDiagonal(twistXform, 1.0f);
	makeHTrans(originXform, mLookAt);
	
	mPosition = Vec3f(0,0,0);
	// grouped for (mat4 * vec3) ops instead of (mat4 * mat4) ops
	mPosition = originXform * (azimXform * (elevXform * (dollyXform * mPosition)));

	if ( fmod(mElevation, 2.0f*M_PI) < 3*M_PI/2 && fmod(mElevation, 2.0f*M_PI) > M_PI/2 )
		mUpVector= Vec3f(0,-1,0);
	else
		mUpVector= Vec3f(0,1,0);

	mDirtyTransform = false;
}

Camera::Camera() 
{
	mElevation = mAzimuth = mTwist = 0.0f;
	mDolly = -20.0f;
	mElevation = 0.2f;
	mAzimuth = (float)M_PI;

	mLookAt = Vec3f( 0, 0, 0 );
	mCurrentMouseAction = kActionNone;

	calculateViewingTransformParameters();
}

void Camera::clickMouse( MouseAction_t action, int x, int y )
{
	mCurrentMouseAction = action;
	mLastMousePosition[0] = x;
	mLastMousePosition[1] = y;
}

void Camera::dragMouse( int x, int y )
{
	Vec3f mouseDelta   = Vec3f(x,y,0.0f) - mLastMousePosition;
	mLastMousePosition = Vec3f(x,y,0.0f);

	switch(mCurrentMouseAction)
	{
	case kActionTranslate:
		{
			calculateViewingTransformParameters();

			double xTrack =  -mouseDelta[0] * kMouseTranslationXSensitivity;
			double yTrack =  mouseDelta[1] * kMouseTranslationYSensitivity;

			Vec3f transXAxis = mUpVector ^ (mPosition - mLookAt);
			transXAxis /= sqrt((transXAxis*transXAxis));
			Vec3f transYAxis = (mPosition - mLookAt) ^ transXAxis;
			transYAxis /= sqrt((transYAxis*transYAxis));

			setLookAt(getLookAt() + transXAxis*xTrack + transYAxis*yTrack);
			
			break;
		}
	case kActionRotate:
		{
			float dAzimuth		=   -mouseDelta[0] * kMouseRotationSensitivity;
			float dElevation	=   mouseDelta[1] * kMouseRotationSensitivity;
			
			setAzimuth(getAzimuth() + dAzimuth);
			setElevation(getElevation() + dElevation);
			
			break;
		}
	case kActionZoom:
		{
			float dDolly = -mouseDelta[1] * kMouseZoomSensitivity;
			setDolly(getDolly() + dDolly);
			break;
		}
	case kActionTwist:
		// Not implemented
	default:
		break;
	}

}

void Camera::releaseMouse( int x, int y )
{
	mCurrentMouseAction = kActionNone;
}

void Camera::fancyPantsCamera( float posX, float posY, float posZ, float lookX, float lookY, float lookZ, float upX, float upY, float upZ) {
	
	float rotX = (lookX-posX);
	float rotY = (lookY-posY);
	float rotZ = (lookZ-posZ);

	float rotMag = sqrt(rotX*rotX + rotY*rotY + rotZ*rotZ);
	rotX = rotX/rotMag;
	rotY = rotY/rotMag;
	rotZ = rotZ/rotMag;

	//float crossX = upY*rotZ-rotY*upZ;
	//float crossY = upZ*rotX-upX*rotZ;
	//float crossZ = upX*rotY-upY*rotX;
	float crossX = rotY*upZ-rotZ*upY;
	float crossY = rotZ*upX-rotX*upZ;
	float crossZ = rotX*upY-rotY*upX;

	float crossMag = sqrt(crossX*crossX + crossY*crossY + crossZ*crossZ);
	crossX = crossX/crossMag;
	crossY = crossY/crossMag;
	crossZ = crossZ/crossMag;
	
	float vX = crossY*rotZ - crossZ*rotY;
	float vY = crossZ*rotX - crossX*rotZ;
	float vZ = crossX*rotY - crossY*rotX;

	//not necessary because cross _|_ rot
	//float vMag = sqrt(vX*vX + vY*vY + vZ*vZ);
	//vX = vX/vMag;
	//vY = vY/vMag;
	//vZ = vZ/vMag;
	
	float per[] = { crossX,vX,-rotX,0,
					crossY,vY,-rotY,0,
					crossZ,vZ,-rotZ,0,
					0,0,0,1 };

	glMultMatrixf( per );
	glTranslatef(-posX, -posY, -posZ);
					
}
void Camera::applyViewingTransform() {
	if( mDirtyTransform )
		calculateViewingTransformParameters();

	// Place the camera at mPosition, aim the camera at
	// mLookAt, and twist the camera such that mUpVector is up
	fancyPantsCamera(	mPosition[0], mPosition[1], mPosition[2],
				mLookAt[0],   mLookAt[1],   mLookAt[2],
				mUpVector[0], mUpVector[1], mUpVector[2]);


	
    // You Will Have to implement this (gluLookAt() ) yourself!
	// what fun that will be!
}

#pragma warning(pop)